/*
 *  
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pvrusb2.h"
#include "pvrusb2-v4l2.h"

#define PVR_SAA7115_I2C_ADDR    0x21

/* These are sequences of commands for the saa7115 chip, in
 * the following format : 0c 21 number_of_cmd 02 cmd data cmd data ... where
 * everything is one byte wise.
 * If you want to tune it or understand what's going on, take a look at
 * saa7175.c from the ivtv project.
 */

PVR_DECLARE_CMD(decoder_set_auto_input, 1,
   "\x0c\x21\x1a\x02\x01\x48\x03\x2C\x04\x90\x05\x90\x06\xeb\x07\xe0"
   "\x0a\x80\x0b\x44\x0c\x40\x0d\x00\x0f\x24\x10\x06\x11\x00\x12\x9d"
   "\x13\x80\x14\x01\x15\x04\x17\x98\x18\x40\x19\x80\x1a\x77\x1b\x42"
   "\x1c\xa9\x1d\x01\x88\xd0\x88\xf0");

PVR_DECLARE_CMD(decoder_reset_scaler, 1,
    "\x0c\x21\x04\x02\x87\x00\x88\x0b\x88\xf0\x87\x01");

PVR_DECLARE_CMD(decoder_set_fullres_x, 1,
    "\x0c\x21\x0e\x02\xcc\xd0\xcd\x02\xd0\x01\xd1\x00\xd2\x00\xd4\x80"
    "\xd5\x40\xd6\x40\xd8\x00\xd9\x04\xda\x00\xdc\x00\xdd\x02\xde\x00");

PVR_DECLARE_CMD(decoder_set_fullres_y_pal, 1,
    "\x0c\x21\x0f\x02\xce\x20\xcf\x01\xe0\x00\xe1\x04\xe2\x00\xe3\x04"
    "\xe4\x01\xe8\x00\xe9\x00\xea\x00\xeb\x00\xec\x00\xed\x00\xee\x00"
    "\xef\x00");

PVR_DECLARE_CMD(decoder_set_fullres_y_ntsc, 1,
    "\x0c\x21\x0f\x02\xce\xfd\xcf\x00\xe0\x00\xe1\x04\xe2\x00\xe3\x04"
    "\xe4\x01\xe8\x00\xe9\x00\xea\x00\xeb\x00\xec\x00\xed\x00\xee\x00"
    "\xef\x00");

PVR_DECLARE_CMD(decoder_set_vcd_pal, 1,
    "\x0c\x21\x1c\x02\xcc\x60\xcd\x01\xce\x20\xcf\x01\xd0\x02\xd8\x18"
    "\xd9\x04\xdc\x0c\xdd\x02\xe0\x00\xe1\x04\xe2\x00\xe3\x04");

PVR_DECLARE_CMD(decoder_set_svcd_pal, 1,
    "\x0c\x21\x07\x02\xce\xe0\xcf\x01\xd0\x01\xd8\x00\xd9\x06\xdc\x00"
    "\xdd\x03");

PVR_DECLARE_CMD(decoder_set_video_pal, 1,
    "\x0c\x21\x1a\x02\x80\x00\x88\x0b\x16\x15\x08\x28\x0e\x07\xc0\x00"
    "\xc1\x08\xc2\x00\xc3\x80\xc4\x00\xc5\x00\xc6\xd0\xc7\x02\xc8\x14"
    "\xc9\x00\xca\x20\xcb\x01\xf0\xb0\xf1\x05\xf5\xb0\xf6\x01\x87\x00"
    "\x88\x0b\x80\xa0\x88\xf0\x87\x01");

PVR_DECLARE_CMD(decoder_set_video_ntsc, 1,
    "\x0c\x21\x1a\x02\x80\x00\x88\x0b\x16\x11\x08\x68\x0e\x07\xc0\x00"
    "\xc1\x08\xc2\x00\xc3\x80\xc4\x02\xc5\x00\xc6\xd0\xc7\x02\xc8\x14"
    "\xc9\x00\xca\xfd\xcb\x00\xf0\xad\xf1\x05\xf5\xad\xf6\x01\x87\x00"
    "\x88\x0b\x80\x20\x88\xf0\x87\x01");

PVR_DECLARE_CMD(decoder_set_misc1, 1,
    "\x0c\x21\x1e\x02\x38\x03\x39\x10\x3a\x00\x81\x01\x82\x00\x83\x01"
    "\x84\x20\x85\x21\x86\xc5\x87\x01\xf2\x50\xf3\x46\xf4\x00\xf7\x4b"
    "\xf8\x00\xf9\x4b\xfa\x00\xfb\x4b\xff\x88\x40\x00\x41\xff\x42\xff"
    "\x43\xff\x44\xff\x45\xff\x46\xff\x47\xff\x48\xff\x49\xff\x4a\xff");

PVR_DECLARE_CMD(decoder_set_misc2, 1,
    "\x0c\x21\x17\x02\x4b\xff\x4c\xff\x4d\xff\x4e\xff\x4f\xff\x50\xff"
    "\x51\xff\x53\xff\x54\xff\x55\xff\x56\xff\x57\xff\x58\x00\x59\x47"
    "\x5a\x06\x5b\x88\x5d\xbf\x5e\x35\x02\x84\x09\x53\x80\x20\x88\xd0"
    "\x88\xf0");

PVR_DECLARE_CMD(decoder_set_audio_48_pal, 1,
    "\x0c\x21\x06\x02\x34\xce\x35\xfb\x36\x30\x30\x00\x31\xc0\x32\x03");

PVR_DECLARE_CMD(decoder_set_audio_44_pal, 1,
    "\x0c\x21\x06\x02\x34\xf2\x35\x00\x36\x2d\x30\x00\x31\x72\x32\x03");

PVR_DECLARE_CMD(decoder_set_audio_32_pal, 1,
    "\x0c\x21\x06\x02\x34\xdf\x35\xa7\x36\x20\x30\x00\x31\x80\x32\x02");

PVR_DECLARE_CMD(decoder_set_audio_48_ntsc, 1,
    "\x0c\x21\x03\x02\x34\xce\x35\xfb\x36\x30\x30\xcd\x31\x20\x32\x03");

PVR_DECLARE_CMD(decoder_set_audio_44_ntsc, 1,
    "\x0c\x21\x03\x02\x34\xf2\x35\x00\x36\x2d\x30\xbc\x31\xdf\x32\x02");

PVR_DECLARE_CMD(decoder_set_audio_32_ntsc, 1,
    "\x0c\x21\x03\x02\x34\xdf\x35\xa7\x36\x20\x30\xde\x31\x15\x32\x02");

PVR_DECLARE_CMD(decoder_vbi_on, 1,
    "\x0c\x21\x05\x02\x80\x00\x88\xd0\x80\x30\x88\xf0\x87\x01");

PVR_DECLARE_CMD(decoder_vbi_off, 1,
    "\x0c\x21\x05\x02\x80\x00\x88\xd0\x80\x20\x88\xf0\x87\x01");


static int pvr_write_decoder_cmd(struct pvr *pvr, u8 cmd, u8 data)
{
    u8 buff[] = { cmd, data };

    pvr_debug("pvr_write_decoder_cmd");

    return pvr_i2c_write(pvr, PVR_SAA7115_I2C_ADDR, buff, 2);
}


int pvr_decoder_set_input(struct pvr *pvr)
{
    switch(pvr->video_input){
        case PVR_INPUT_TV:
            return pvr_write_decoder_cmd(pvr, 0x02, 0x84);
        case PVR_INPUT_COMPOSITE:
            return pvr_write_decoder_cmd(pvr, 0x02, 0x85);
        case PVR_INPUT_SVIDEO:
            return pvr_write_decoder_cmd(pvr, 0x02, 0x88);
    }
    return -EINVAL;
}

int pvr_decoder_set_bcsh(struct pvr *pvr)
{
    int ret = 0;

    ret |= pvr_write_decoder_cmd(pvr, 0x0a, pvr->video_brightness);
    ret |= pvr_write_decoder_cmd(pvr, 0x0b, pvr->video_contrast);
    ret |= pvr_write_decoder_cmd(pvr, 0x0c, pvr->video_saturation);
    ret |= pvr_write_decoder_cmd(pvr, 0x0d, (s8) pvr->video_hue);

    return ret;
}

int pvr_configure_decoder(struct pvr *pvr)
{
    int ret = 0;

    /* Set some default values then set the right resolution. */
    
    ret |= pvr_do_command(pvr, &decoder_set_auto_input);
    ret |= pvr_do_command(pvr, &decoder_set_misc1);
    ret |= pvr_do_command(pvr, &decoder_set_misc2);
    ret |= pvr_do_command(pvr, &decoder_set_fullres_x);
    
    if (pvr->standard == PVR_STD_NTSC){
        ret |= pvr_do_command(pvr, &decoder_set_fullres_y_ntsc);
        ret |= pvr_do_command(pvr, &decoder_set_video_ntsc);
        switch(pvr->audio_srate){
            case PVR_SRATE_48:
                ret |= pvr_do_command(pvr, &decoder_set_audio_48_ntsc);
                break;
            case PVR_SRATE_44:
                ret |= pvr_do_command(pvr, &decoder_set_audio_44_ntsc);
                break;
        }
    } else {
        ret |= pvr_do_command(pvr, &decoder_set_fullres_y_pal);
        ret |= pvr_do_command(pvr, &decoder_set_video_pal);
        switch(pvr->audio_srate){
            case PVR_SRATE_48:
                ret |= pvr_do_command(pvr, &decoder_set_audio_48_pal);
                break;
            case PVR_SRATE_44:
                ret |= pvr_do_command(pvr, &decoder_set_audio_44_pal);
                break;
        }
    }

    if (pvr->video_width == PVR_WIDTH_VCD)
        ret |= pvr_do_command(pvr, &decoder_set_vcd_pal);
    else if (pvr->video_width == PVR_WIDTH_SVCD)
        ret |= pvr_do_command(pvr, &decoder_set_svcd_pal);

    ret |= pvr_do_command(pvr, &decoder_reset_scaler);

    return ret;
}
