/*
 *  
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pvrusb2.h"
#include "pvrusb2-v4l2.h"

#define PVR_TUNER_I2C_ADDR  0x61
#define PVR_STD_I2C_ADDR    0x43

/*
 * pvr_set_freq()
 *
 * Set the frequency to freq (unit is 62.5 KHz).
 *
 */

struct pvr_tuner 
{
	char *name;
	unsigned char std;
  
	unsigned short thresh1; /* limit between VHF LOW and VHF HIGH */
	unsigned short thresh2; /* limit between VHF HIGH and UHF */
	unsigned char vhf_l;
	unsigned char vhf_h;
	unsigned char uhf;
	unsigned char config;
	unsigned short IFPCoff; /* 622.4=16*38.90 MHz PAL, 
				   732  =16*45.75 NTSCi, 
				   940  =58.75 NTSC-Japan */
};

/*
 * Tuner found in pvrusb2 devices. As the microtune does only PAL-BGIDK and
 * SECAM, there should be at least 3 or 4 different tuner around.
 * NOTE : This data is not the same as kernel tuner.c one. Why ?
 */

#define PVR_NTUNER  1

static struct pvr_tuner pvr_tuners[] = {
    { "Microtune 4049 FM5", 0 /* PAL-BGDKI SECAM */,
      16*141, 16*464, 0xa0, 0x90, 0x30, 0xce, 623 },

};

int pvr_set_standard(struct pvr *pvr)
{
    u8 buff [4]; 

    pvr_debug("pvr_set_standard");

    PVR_BUG_ON(pvr->standard < 0 || pvr->standard >= PVR_V4L2_NSTD);

    /* FIXME: Might be tuner dependant ? Need more hardware info... */
    PVR_DECOMPOSE_BE(buff, 0, pvr_tbl_standard[pvr->standard]);

    return pvr_i2c_write(pvr, PVR_STD_I2C_ADDR, buff, 4);
}

int pvr_set_freq(struct pvr *pvr)
{
    u32 val;
    u8 buff[4];

    pvr_debug("pvr_set_freq : %i", pvr->frequency);

    PVR_BUG_ON(pvr->tuner < 0 || pvr->tuner >= PVR_NTUNER);

    val = pvr_tuners[pvr->tuner].IFPCoff + pvr->frequency;

    buff[0] = val >> 8;
    buff[1] = val & 0xff;
    buff[2] = pvr_tuners[pvr->tuner].config;

    if (pvr->frequency < pvr_tuners[pvr->tuner].thresh1)
        /* VHF LOW */
        buff[3] = pvr_tuners[pvr->tuner].vhf_l;
    else if (pvr->frequency < pvr_tuners[pvr->tuner].thresh2) 
        /* VHF HIGH */
        buff[3] = pvr_tuners[pvr->tuner].vhf_h;
    else
        /* UHF */
        buff[3] = pvr_tuners[pvr->tuner].uhf;

    return pvr_i2c_write(pvr, PVR_TUNER_I2C_ADDR, buff, 4);
}
